@@ -40,25 +40,48 @@
 #include "filesdb.h"
 #include "main.h"
 
-struct pkginqueue {
-  struct pkginqueue *next;
-  struct pkginfo *pkg;
-};
+static struct pkginfo *progress_bytrigproc;
+static PKGQUEUE_DEF_INIT(queue);
 
-static struct pkginqueue *queuehead= 0, **queuetail= &queuehead;
+int sincenothing = 0, dependtry = 0;
 
-int queuelen=0, sincenothing=0, dependtry=0;
-
-void add_to_queue(struct pkginfo *pkg) {
+struct pkginqueue *
+add_to_some_queue(struct pkginfo *pkg, struct pkgqueue *q)
+{
   struct pkginqueue *newent;
 
   newent= m_malloc(sizeof(struct pkginqueue));
   newent->pkg= pkg;
-  newent->next= 0;
-  *queuetail= newent;
-  queuetail= &newent->next;
+  newent->next = NULL;
+  *q->tail = newent;
+  q->tail = &newent->next;
+  q->length++;
 
-  queuelen++;
+  return newent;
+}
+
+struct pkginqueue *
+remove_from_some_queue(struct pkgqueue *q)
+{
+  struct pkginqueue *removeent = q->head;
+
+  if (!removeent)
+    return NULL;
+
+  assert(q->length > 0);
+
+  q->head = q->head->next;
+  if (q->tail == &removeent->next)
+    q->tail= &q->head;
+  q->length--;
+
+  return removeent;
+}
+
+void
+add_to_queue(struct pkginfo *pkg)
+{
+  add_to_some_queue(pkg, &queue);
 }
 
 void packages(const char *const *argv) {
@@ -67,11 +90,14 @@ void packages(const char *const *argv) {
   const char *thisarg;
   size_t l;
   
+  trigproc_install_hooks();
+
   modstatdb_init(admindir,
                  f_noact ?    msdbrw_readonly
                : fc_nonroot ? msdbrw_write
                :              msdbrw_needsuperuser);
   checkpath();
+  log_message("startup packages %s", cipaction->olong);
 
   if (f_pending) {
 
@@ -79,10 +105,18 @@ void packages(const char *const *argv) {
       badusage(_("--%s --pending does not take any non-option arguments"),cipaction->olong);
 
     it= iterpkgstart();
-    while ((pkg= iterpkgnext(it)) != 0) {
+    while ((pkg = iterpkgnext(it)) != NULL) {
       switch (cipaction->arg) {
       case act_configure:
-        if (pkg->status != stat_unpacked && pkg->status != stat_halfconfigured)
+        if (!(pkg->status == stat_unpacked ||
+              pkg->status == stat_halfconfigured ||
+              pkg->trigpend_head))
+          continue;
+        if (pkg->want != want_install)
+          continue;
+        break;
+      case act_triggers:
+        if (!pkg->trigpend_head)
           continue;
         if (pkg->want != want_install)
           continue;
@@ -108,7 +142,7 @@ void packages(const char *const *argv) {
     if (!*argv)
       badusage(_("--%s needs at least one package name argument"), cipaction->olong);
     
-    while ((thisarg= *argv++) != 0) {
+    while ((thisarg = *argv++) != NULL) {
       pkg= findpackage(thisarg);
       if (pkg->status == stat_notinstalled) {
         l= strlen(pkg->name);
@@ -124,6 +158,7 @@ void packages(const char *const *argv) {
   ensure_diversions();
 
   process_queue();
+  trigproc_run_deferred();
 
   modstatdb_shutdown();
 }
@@ -131,21 +166,27 @@ void packages(const char *const *argv) {
 void process_queue(void) {
   struct pkginqueue *removeent, *rundown;
   struct pkginfo *volatile pkg;
+  enum action action_todo;
   jmp_buf ejbuf;
   enum istobes istobe= itb_normal;
   
+  if (abort_processing)
+    return;
+
   clear_istobes();
 
   switch (cipaction->arg) {
+  case act_triggers:
   case act_configure: case act_install:  istobe= itb_installnew;  break;
   case act_remove: case act_purge:       istobe= itb_remove;      break;
   default: internerr("unknown action for queue start");
   }
-  for (rundown= queuehead; rundown; rundown= rundown->next) {
+  for (rundown = queue.head; rundown; rundown = rundown->next) {
     ensure_package_clientdata(rundown->pkg);
     if (rundown->pkg->clientdata->istobe == istobe) {
-      /* Remove it from the queue - this is a second copy ! */
+      /* Erase the queue entry - this is a second copy! */
       switch (cipaction->arg) {
+      case act_triggers:
       case act_configure: case act_remove: case act_purge:
         printf(_("Package %s listed more than once, only processing once.\n"),
                rundown->pkg->name);
@@ -158,45 +199,61 @@ void process_queue(void) {
       default:
         internerr("unknown action in duplicate");
       }
-      rundown->pkg= 0;
+      rundown->pkg = NULL;
    } else {
       rundown->pkg->clientdata->istobe= istobe;
     }
   }
   
-  while (queuelen) {
-    removeent= queuehead;
-    assert(removeent);
-    queuehead= queuehead->next;
-    queuelen--;
-    if (queuetail == &removeent->next) queuetail= &queuehead;
-
+  while ((removeent = remove_from_some_queue(&queue))) {
     pkg= removeent->pkg;
     free(removeent);
     
     if (!pkg) continue; /* duplicate, which we removed earlier */
 
-    assert(pkg->status <= stat_configfiles);
+    action_todo = cipaction->arg;
+
+    if (sincenothing++ > queue.length * 2 + 2) {
+      if (progress_bytrigproc && progress_bytrigproc->trigpend_head) {
+        add_to_queue(pkg);
+        pkg = progress_bytrigproc;
+        action_todo = act_configure;
+      } else {
+        dependtry++;
+        sincenothing = 0;
+        assert(dependtry <= 4);
+      }
+    }
+
+    assert(pkg->status <= stat_installed);
 
     if (setjmp(ejbuf)) {
       /* give up on it from the point of view of other packages, ie reset istobe */
       pkg->clientdata->istobe= itb_normal;
       error_unwind(ehflag_bombout);
-      if (onerr_abort > 0) break;
+      if (abort_processing)
+        return;
       continue;
     }
     push_error_handler(&ejbuf,print_error_perpackage,pkg->name);
-    if (sincenothing++ > queuelen*2+2) {
-      dependtry++; sincenothing= 0;
-      assert(dependtry <= 4);
-    }
-    switch (cipaction->arg) {
+    switch (action_todo) {
+    case act_triggers:
+      if (!pkg->trigpend_head)
+        ohshit(_("package %.250s is not ready for trigger processing\n"
+                 " (current status `%.250s' with no pending triggers)"),
+               pkg->name, statusinfos[pkg->status].name);
+      /* Fall through. */
     case act_install:
       /* Don't try to configure pkgs that we've just disappeared. */
       if (pkg->status == stat_notinstalled)
         break;
+      /* Fall through. */
     case act_configure:
-      deferred_configure(pkg);
+      /* Do whatever is most needed. */
+      if (pkg->trigpend_head)
+        trigproc(pkg);
+      else
+        deferred_configure(pkg);
       break;
     case act_remove: case act_purge:
       deferred_remove(pkg);
@@ -206,9 +263,10 @@ void process_queue(void) {
     }
     if (ferror(stdout)) werr("stdout");
     if (ferror(stderr)) werr("stderr");
-    set_error_display(0,0);
+    set_error_display(NULL, NULL);
     error_unwind(ehflag_normaltidy);
   }
+  assert(!queue.length);
 }    
 
 /*** dependency processing - common to --configure and --remove ***/
@@ -248,10 +306,21 @@ void process_queue(void) {
  * and breaking.
  */
 
+/* Return values:
+ *   0: cannot be satisfied.
+ *   1: defer: may be satisfied later, when other packages are better or
+ *      at higher dependtry due to --force
+ *      will set *fixbytrig to package whose trigger processing would help
+ *      if applicable (and leave it alone otherwise)
+ *   2: not satisfied but forcing
+ *      (*interestingwarnings >= 0 on exit? caller is to print oemsgs)
+ *   3: satisfied now
+ */
 static int deppossi_ok_found(struct pkginfo *possdependee,
                              struct pkginfo *requiredby,
                              struct pkginfo *removing,
                              struct pkginfo *providing,
+                             struct pkginfo **fixbytrig,
                              int *matched,
                              struct deppossi *checkversion,
                              int *interestingwarnings,
@@ -279,9 +348,11 @@ static int deppossi_ok_found(struct pkginfo *possdependee,
     return thisf;
   }
   switch (possdependee->status) {
-  case stat_installed:
   case stat_unpacked:
   case stat_halfconfigured:
+  case stat_triggersawaited:
+  case stat_triggerspending:
+  case stat_installed:
     assert(possdependee->installed.valid);
     if (checkversion && !versionsatisfied(&possdependee->installed,checkversion)) {
       varbufprintf(oemsgs, _("  Version of %s on system is %s.\n"),
@@ -294,15 +365,51 @@ static int deppossi_ok_found(struct pkginfo *possdependee,
       (*interestingwarnings)++;
       return thisf;
     }
-    if (possdependee->status == stat_installed) {
+    if (possdependee->status == stat_installed ||
+        possdependee->status == stat_triggerspending) {
       debug(dbg_depcondetail,"      is installed, ok and found");
       return 3;
     }
+    if (possdependee->status == stat_triggersawaited) {
+      assert(possdependee->trigaw.head);
+      if (removing || !(f_triggers ||
+                        possdependee->clientdata->istobe == itb_installnew)) {
+        if (providing) {
+          varbufprintf(oemsgs,
+                       _("  Package %s which provides %s awaits trigger processing.\n"),
+                       possdependee->name, providing->name);
+        } else {
+          varbufprintf(oemsgs,
+                       _("  Package %s awaits trigger processing.\n"),
+                       possdependee->name);
+        }
+        debug(dbg_depcondetail, "      triggers-awaited, no fixbytrig");
+        goto unsuitable;
+      }
+      /* We don't check the status of trigaw.head->pend here, just in case
+       * we get into the pathological situation where Triggers-Awaited but
+       * the named package doesn't actually have any pending triggers. In
+       * that case we queue the non-pending package for trigger processing
+       * anyway, and that trigger processing will be a noop except for
+       * sorting out all of the packages which name it in T-Awaited.
+       *
+       * (This situation can only arise if modstatdb_note success in
+       * clearing the triggers-pending status of the pending package
+       * but then fails to go on to update the awaiters.)
+       */
+      *fixbytrig = possdependee->trigaw.head->pend;
+      debug(dbg_depcondetail,
+            "      triggers-awaited, fixbytrig `%s', returning 1",
+            (*fixbytrig)->name);
+      return 1;
+    }
     if (possdependee->clientdata &&
         possdependee->clientdata->istobe == itb_installnew) {
       debug(dbg_depcondetail,"      unpacked/halfconfigured, defer");
       return 1;
-    } else if (!removing && fc_configureany && !skip_due_to_hold(possdependee)) {
+    } else if (!removing && fc_configureany &&
+               !skip_due_to_hold(possdependee) &&
+               !(possdependee->status == stat_halfconfigured)) {
       fprintf(stderr,
               _("dpkg: also configuring `%s' (required by `%s')\n"),
               possdependee->name, requiredby->name);
@@ -317,11 +424,10 @@ static int deppossi_ok_found(struct pkginfo *possdependee,
 		     possdependee->name);
       }
 
-      if (fc_depends) thisf= (dependtry >= 4) ? 2 : 1;
-      debug(dbg_depcondetail,"      not configured/able - returning %d",thisf);
-      (*interestingwarnings)++;
-      return thisf;
+      debug(dbg_depcondetail, "      not configured/able");
+      goto unsuitable;
     }
+
   default:
     if (providing) {
       varbufprintf(oemsgs,
@@ -332,19 +438,100 @@ static int deppossi_ok_found(struct pkginfo *possdependee,
 		   possdependee->name);
     }
 
-    if (fc_depends) thisf= (dependtry >= 4) ? 2 : 1;
-    debug(dbg_depcondetail,"      not installed - returning %d",thisf);
-    (*interestingwarnings)++;
-    return thisf;
+    debug(dbg_depcondetail, "      not installed");
+    goto unsuitable;
   }
+
+unsuitable:
+  if (fc_depends)
+    thisf = (dependtry >= 4) ? 2 : 1;
+
+  debug(dbg_depcondetail, "        returning %d", thisf);
+  (*interestingwarnings)++;
+
+  return thisf;
+}
+
+static void breaks_check_one(struct varbuf *aemsgs, int *ok,
+                             struct deppossi *breaks,
+                             struct pkginfo *broken,
+                             struct pkginfo *breaker,
+                             struct pkginfo *virtbroken) {
+  struct varbuf depmsg;
+
+  debug(dbg_depcondetail, "      checking breaker %s virtbroken %s",
+        breaker->name, virtbroken ? virtbroken->name : "<none>");
+
+  if (breaker->status == stat_notinstalled ||
+      breaker->status == stat_configfiles) return;
+  if (broken == breaker) return;
+  if (!versionsatisfied(&broken->installed, breaks)) return;
+  if (ignore_depends(breaker)) return;
+  if (virtbroken && ignore_depends(virtbroken)) return;
+
+  varbufinit(&depmsg);
+  varbufdependency(&depmsg, breaks->up);
+  varbufaddc(&depmsg, 0);
+  varbufprintf(aemsgs, _(" %s (%s) breaks %s and is %s.\n"),
+               breaker->name,
+               versiondescribe(&breaker->installed.version, vdew_nonambig),
+               depmsg.buf,
+               gettext(statusstrings[breaker->status]));
+  varbuffree(&depmsg);
+
+  if (virtbroken) {
+    varbufprintf(aemsgs, _("  %s (%s) provides %s.\n"),
+                 broken->name,
+                 versiondescribe(&broken->installed.version, vdew_nonambig),
+                 virtbroken->name);
+  } else if (breaks->verrel != dvr_none) {
+    varbufprintf(aemsgs, _("  Version of %s to be configured is %s.\n"),
+                 broken->name,
+                 versiondescribe(&broken->installed.version, vdew_nonambig));
+    if (fc_dependsversion) return;
+  }
+  if (force_breaks(breaks)) return;
+  *ok= 0;
+}
+
+static void breaks_check_target(struct varbuf *aemsgs, int *ok,
+                                struct pkginfo *broken,
+                                struct pkginfo *target,
+                                struct pkginfo *virtbroken) {
+  struct deppossi *possi;
+
+  for (possi= target->installed.depended; possi; possi= possi->nextrev) {
+    if (possi->up->type != dep_breaks) continue;
+    if (virtbroken && possi->verrel != dvr_none) continue;
+    breaks_check_one(aemsgs, ok, possi, broken, possi->up->up, virtbroken);
+  }
+}
+
+int breakses_ok(struct pkginfo *pkg, struct varbuf *aemsgs) {
+  struct dependency *dep;
+  struct pkginfo *virtbroken;
+  int ok= 2;
+
+  debug(dbg_depcon, "    checking Breaks");
+
+  breaks_check_target(aemsgs, &ok, pkg, pkg, NULL);
+
+  for (dep= pkg->installed.depends; dep; dep= dep->next) {
+    if (dep->type != dep_provides) continue;
+    virtbroken= dep->list->ed;
+    debug(dbg_depcondetail, "     checking virtbroken %s", virtbroken->name);
+    breaks_check_target(aemsgs, &ok, pkg, virtbroken, virtbroken);
+  }
+  return ok;
 }
 
 int dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
                     struct varbuf *aemsgs) {
-  int ok, matched, found, thisf, interestingwarnings;
+  int ok, matched, found, thisf, interestingwarnings, anycannotfixbytrig;
   struct varbuf oemsgs;
   struct dependency *dep;
   struct deppossi *possi, *provider;
+  struct pkginfo *possfixbytrig, *canfixbytrig;
 
   varbufinit(&oemsgs);
   interestingwarnings= 0;
@@ -352,21 +539,23 @@ int dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
   debug(dbg_depcon,"checking dependencies of %s (- %s)",
         pkg->name, removing ? removing->name : "<none>");
   assert(pkg->installed.valid);
-  /* To implement Breaks we need to add code here which prevents
-   * configuration of Broken packages.
-   * FIXME: implement Breaks */
+
+  anycannotfixbytrig = 0;
+  canfixbytrig = NULL;
   for (dep= pkg->installed.depends; dep; dep= dep->next) {
     if (dep->type != dep_depends && dep->type != dep_predepends) continue;
     debug(dbg_depcondetail,"  checking group ...");
     matched= 0; varbufreset(&oemsgs);
     found= 0; /* 0=none, 1=defer, 2=withwarning, 3=ok */
+    possfixbytrig = NULL;
     for (possi= dep->list; found != 3 && possi; possi= possi->next) {
       debug(dbg_depcondetail,"    checking possibility  -> %s",possi->ed->name);
       if (possi->cyclebreak) {
         debug(dbg_depcondetail,"      break cycle so ok and found");
         found= 3; break;
       }
-      thisf= deppossi_ok_found(possi->ed,pkg,removing,0,
+      thisf = deppossi_ok_found(possi->ed, pkg, removing, NULL,
+                                &possfixbytrig,
                                &matched,possi,&interestingwarnings,&oemsgs);
       if (thisf > found) found= thisf;
       if (found != 3 && possi->verrel == dvr_none) {
@@ -377,7 +566,8 @@ int dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
             if (provider->up->type != dep_provides) continue;
             debug(dbg_depcondetail,"     checking provider %s",provider->up->up->name);
             thisf= deppossi_ok_found(provider->up->up,pkg,removing,possi->ed,
-                                     &matched,0,&interestingwarnings,&oemsgs);
+                                     &possfixbytrig,
+                                     &matched, NULL, &interestingwarnings, &oemsgs);
             if (thisf > found) found= thisf;
           }
         }
@@ -385,10 +575,12 @@ int dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
       debug(dbg_depcondetail,"    found %d",found);
       if (thisf > found) found= thisf;
     }
-    debug(dbg_depcondetail,"  found %d matched %d",found,matched);
+    debug(dbg_depcondetail, "  found %d matched %d possfixbytrig %s",
+          found, matched, possfixbytrig ? possfixbytrig->name : "-");
     if (removing && !matched) continue;
     switch (found) {
     case 0:
+      anycannotfixbytrig = 1;
       ok= 0;
     case 2:
       varbufaddstr(aemsgs, " ");
@@ -407,6 +599,10 @@ int dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
       }
       break;
     case 1:
+      if (possfixbytrig)
+        canfixbytrig = possfixbytrig;
+      else
+        anycannotfixbytrig = 1;
       if (ok>1) ok= 1;
       break;
     case 3:
@@ -417,6 +613,8 @@ int dependencies_ok(struct pkginfo *pkg, struct pkginfo *removing,
   }
   if (ok == 0 && (pkg->clientdata && pkg->clientdata->istobe == itb_remove))
     ok= 1;
+  if (!anycannotfixbytrig && canfixbytrig)
+    progress_bytrigproc = canfixbytrig;
   
   varbuffree(&oemsgs);
   debug(dbg_depcon,"ok %d msgs >>%.*s<<", ok, (int)aemsgs->used, aemsgs->buf);
